c ======================================================================
c $Id: DynTest.txt,v 1.1 2002/04/18 21:24:17 MarioStorti Exp apache $
c
c Updated and cleaned source code to the 2010 year (jdelia 2010-10-18):
c
c  1) The second() function replaced by the intrinsic cpu_time subroutine.
c  2) Implicit conversion were removed.
c  3) Nonconforming tab characters were removed.
c  4) The compilation is OK with the following compilers:
c       gfortran: reasonable performance.
c       ifort   : reasonable performance
c       g95     : bad performance in 64 (compiler flags should be improved)
c  5) Compiler options: see Makefile file.
c
c ----------------------------------------------------------------------
        program lbn
        implicit real*8 (a-h,o-z) 
        parameter (lmax=10000000)
        parameter (nvi=20,nvf=1000)
        dimension s(lmax)
c
        iend = 1
        idebug = 0
c  
        n = nvi
        do
          if (n .gt. nvf) exit
          write (*,*) 'n: ',n
          call aloca (ia   ,(2*n+1)*2*n  ,iend,lmax,'a'    ,idebug)
          call aloca (ib   , 2*n         ,iend,lmax,'n'    ,idebug)
          call aloca (ix   , 2*n         ,iend,lmax,'x'    ,idebug)
          call aloca (iacpy, (2*n+1)*2*n ,iend,lmax,'acpy' ,idebug)
          call aloca (ibcpy,  2*n        ,iend,lmax,'bcpy' ,idebug)
          call aloca (iivpt,    n        ,iend,lmax,'ivpt' ,idebug)
          call doit  (n,s(ia),s(ib),s(iacpy),s(ibcpy))
c         free memory
          iend = 1
          n = int (1.3*n)
        end do 
c
        stop
        end program
c ======================================================================
c This program was updated on 10/12/92 to correct a
c problem with the random number generator. The previous
c random number generator had a short period and produced
c singular matrices occasionally.
c ----------------------------------------------------------------------
        subroutine doit (n,a,b,acpy,bcpy)
        integer          ipvt(2*n)
        double precision a(2*n+1,2*n),b(2*n)
        double precision acpy(2*n+1,2*n),bcpy(2*n),ttot(3)
        double precision ops,norma
        double precision t0, t0_3, dev, err, maxerr
c       double precision cray ! variable 'cray' is set but never used
        lda = 2*n+1
c       cray = 0.056
        ops = (2.0d0*dfloat(n)**3)/3.0d0 + 2.0d0*dfloat(n)**2
c
        call matgen (acpy,lda,n,bcpy,norma)
c
c->	open(7,name='lb100dl.out')
c->	do j=1,n
c->	  do i=1,n
c->	    write (7,*) acpy(i,j)
c->	  enddo
c->	enddo
c->	do i=1,n
c->	  write(7,*) bcpy(i)
c->	enddo
c
        niter = 200*100**3/n**3
        if (niter.lt.5) niter=5
        write(*,*) 'niter: ', niter
        do icptim=1,3
c         t1 = second()
          call cpu_time (t1)
          do k=1,niter
            do jcp=icptim,1,-1
              do kk=1,n
                do ll=1,n
                  a(kk,ll) = k*jcp*acpy(kk,ll)
                end do
                b(kk)=k*jcp*bcpy(kk)
              end do
            end do
            call dgefa(a,lda,n,ipvt,info)
            if (info.ne.0) then
              write(*,*) 'On iter ',k,' dgefa returned info= ',info
              stop
            end if
            call dgesl(a,lda,n,ipvt,b,0)
          end do
!         t2=second()
          call cpu_time (t2)
          ttot(icptim)=t2-t1
          write(*,*) 'On iter ',icptim,' ttot: ',ttot(icptim)
        enddo
        t0 = 2 * ttot (1) - ttot(2)
        t0_3 = 1.5*ttot(1)-0.5*ttot(3)
        dev = abs(t0_3-t0)/t0
        if (dev.gt.0.1) then
          write (*,*) 'warning: max dev. exceeded: '
          write (*,*) 't0= ',t0,' t0_3= ',t0_3,' relative dev= ',dev
        end if
        write(*,*) 'Elapsed: ',t0,' per iter: ',t0/niter,
     *    ' rate: ',niter*ops/(1e6*t0),' [mflops]'
c       Verify
        err=0.0d0
        maxerr=0.0d0
        do k=1,n
          err = abs(b(k)-1.0d0)
          if (err .gt. maxerr) maxerr = err
        enddo
        if (err.gt.1e-10) write (*,*) 'warning: max err: ',maxerr
c->     do i=1,n
c->       write (7,*) b(i)
c->     enddo
c->     close(7)
c
        return
        end subroutine
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        subroutine matgen(a,lda,n,b,norma)
        integer lda,n,init(4),i,j
        double precision a(lda,1),b(1),norma,ran
        external ran
c
        init (1) = 1
        init (2) = 2
        init (3) = 3
        init (4) = 1325
        norma = 0.0
        do 30 j = 1,n
          do 20 i = 1,n
            a(i,j) = ran (init) - 0.5
            norma = dmax1(dabs(a(i,j)), norma)
   20     continue
   30   continue
        do 35 i = 1,n
          b(i) = 0.0
   35   continue
        do 50 j = 1,n
          do 40 i = 1,n
            b(i) = b(i) + a(i,j)
   40     continue
   50   continue
c
        return 
        end subroutine
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c
c  dgefa factors a double precision matrix by gaussian elimination.
c
c  dgefa is usually called by dgeco, but it can be called
c  directly with a saving in time if  rcond  is not needed.
c  (time for dgeco) = (1 + 9/n)*(time for dgefa) .
c
c  on entry
c
c	  a	double precision(lda, n)
c		the matrix to be factored.
c
c	  lda	integer
c		the leading dimension of the array  a .
c
c	  n	integer
c		the order of the matrix  a .
c
c  on return
c
c	  a	 an upper triangular matrix and the multipliers
c		 which were used to obtain it.
c		 the factorization can be written  a = l*u  where
c		 l  is a product of permutation and unit lower
c		 triangular matrices and  u  is upper triangular.
c
c	  ipvt	 integer(n)
c		 an integer vector of pivot indices.
c
c	  info	 integer
c                = 0  normal value.
c                = k  if  u(k,k) .eq. 0.0 .  this is not an error
c                condition for this subroutine, but it does
c                indicate that dgesl or dgedi will divide by zero
c                if called.  use  rcond  in dgeco for a reliable
c                indication of singularity.
c
c         linpack. this version dated 08/14/78 .
c         cleve moler, university of new mexico, argonne national lab.
c
c         subroutines and functions
c
c         blas daxpy,dscal,idamax
c ----------------------------------------------------------------------
        subroutine dgefa(a,lda,n,ipvt,info)
        integer lda,n,ipvt(1),info
        double precision a(lda,1)
c       internal variables
        double precision t
        integer idamax,j,k,kp1,l,nm1
c
c       gaussian elimination with partial pivoting
c
        info = 0
        nm1 = n - 1
        if (nm1 .lt. 1) go to 70
        do 60 k = 1, nm1
          kp1 = k + 1
c
c         find l = pivot index
c
          l = idamax(n-k+1,a(k,k),1) + k - 1
          ipvt(k) = l
c
c         zero pivot implies this column already triangularized
c
          if (a(l,k) .eq. 0.0d0) go to 40
c
c         interchange if necessary
c
          if (l .eq. k) go to 10
          t = a(l,k)
          a(l,k) = a(k,k)
          a(k,k) = t
   10     continue
c
c       compute multipliers
c
        t = -1.0d0/a(k,k)
        call dscal(n-k,t,a(k+1,k),1)
c
c       row elimination with column indexing
c
        do 30 j = kp1, n
          t = a(l,j)
          if (l .eq. k) go to 20
          a(l,j) = a(k,j)
          a(k,j) = t
   20     continue
          call daxpy (n-k,t,a(k+1,k),1,a(k+1,j),1)
   30   continue
        go to 50
   40   continue
        info = k
   50   continue
   60   continue
   70   continue
        ipvt(n) = n
        if (a(n,n) .eq. 0.0d0) info = n
c
        return
        end subroutine
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c
c  dgesl solves the double precision system
c  a * x = b  or  trans(a) * x = b
c  using the factors computed by dgeco or dgefa.
c
c  on entry
c
c	  a       double precision(lda, n)
c                  the output from dgeco or dgefa.
c
c         lda     integer
c                 the leading dimension of the array  a .
c
c         n       integer
c                 the order of the matrix  a .
c
c         ipvt    integer(n)
c                 the pivot vector from dgeco or dgefa.
c
c         b       double precision(n)
c                 the right hand side vector.
c
c         job     integer
c          = 0	     to solve  a*x = b ,
c          = nonzero to solve  trans(a)*x = b  where
c            trans(a)  is the transpose.
c
c	  on return
c
c         b       the solution vector  x .
c
c         error condition
c
c  a division by zero will occur if the input factor contains a
c  zero on the diagonal.  technically this indicates singularity
c  but it is often caused by improper arguments or improper
c  setting of lda .  it will not occur if the subroutines are
c  called correctly and if dgeco has set rcond .gt. 0.0
c  or dgefa has set info .eq. 0 .
c
c  to compute  inverse(a) * c  where  c  is a matrix
c  with  p  columns
c  call dgeco(a,lda,n,ipvt,rcond,z)
c  if (rcond is too small) go to ...
c  do 10 j = 1, p
c  call dgesl(a,lda,n,ipvt,c(1,j),0)
c  10 continue
c
c  linpack. this version dated 08/14/78 .
c  cleve moler, university of new mexico, argonne national lab.
c
c  subroutines and functions
c  blas daxpy,ddot
c ----------------------------------------------------------------------
        subroutine dgesl(a,lda,n,ipvt,b,job)
        integer lda,n,ipvt(1),job
        double precision a(lda,1),b(1)
c
c       internal variables
c
        double precision ddot,t
        integer k,kb,l,nm1
c
        nm1 = n - 1
        if (job .ne. 0) go to 50
c
c       job = 0 , solve  a * x = b
c       first solve  l*y = b
c
        if (nm1 .lt. 1) go to 30
        do 20 k = 1, nm1
          l = ipvt(k)
          t = b(l)
          if (l .eq. k) go to 10
          b(l) = b(k)
          b(k) = t
   10     continue
          call daxpy(n-k,t,a(k+1,k),1,b(k+1),1)
   20   continue
   30   continue
c
c	now solve  u*x = y
c
        do 40 kb = 1, n
          k = n + 1 - kb
          b(k) = b(k)/a(k,k)
          t = -b(k)
          call daxpy(k-1,t,a(1,k),1,b(1),1)
   40   continue
        go to 100
   50   continue
c
c       job = nonzero, solve  trans(a) * x = b
c       first solve  trans(u)*y = b
c
        do 60 k = 1, n
          t = ddot(k-1,a(1,k),1,b(1),1)
          b(k) = (b(k) - t)/a(k,k)
   60   continue
c
c       now solve trans(l)*x = y
c
        if (nm1 .lt. 1) go to 90
        do 80 kb = 1, nm1
          k = n - kb
          b(k) = b(k) + ddot(n-k,a(k+1,k),1,b(k+1),1)
          l = ipvt(k)
          if (l .eq. k) go to 70
          t = b(l)
          b(l) = b(k)
          b(k) = t
   70   continue
   80   continue
   90   continue
  100   continue
c
        return
        end subroutine
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c Constant times a vector plus a vector.
c jack dongarra, linpack, 3/11/78.
c ----------------------------------------------------------------------
        subroutine daxpy(n,da,dx,incx,dy,incy)
        double precision dx(1),dy(1),da
        integer i,incx,incy,ix,iy,n
c
        if(n.le.0)return
        if (da .eq. 0.0d0) return
        if(incx.eq.1.and.incy.eq.1)go to 20
c
c	code for unequal increments or equal increments
c	not equal to 1
c
        ix = 1
        iy = 1
        if(incx.lt.0)ix = (-n+1)*incx + 1
        if(incy.lt.0)iy = (-n+1)*incy + 1
        do 10 i = 1,n
          dy(iy) = dy(iy) + da*dx(ix)
          ix = ix + incx
          iy = iy + incy
   10   continue
        return
c
c       code for both increments equal to 1
c
   20   continue
        do 30 i = 1,n
          dy(i) = dy(i) + da*dx(i)
   30   continue
c
        return
        end subroutine
c +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c  Forms the dot product of two vectors.
c  jack dongarra, linpack, 3/11/78.
c ----------------------------------------------------------------------
        double precision function ddot(n,dx,incx,dy,incy)
        double precision dx(1),dy(1),dtemp
        integer i,incx,incy,ix,iy,n
c
        ddot = 0.0d0
        dtemp = 0.0d0
        if(n.le.0)return
        if(incx.eq.1.and.incy.eq.1)go to 20
c
c       code for unequal increments or equal increments
c       not equal to 1
c
        ix = 1
        iy = 1
        if(incx.lt.0)ix = (-n+1)*incx + 1
        if(incy.lt.0)iy = (-n+1)*incy + 1
        do 10 i = 1,n
          dtemp = dtemp + dx(ix)*dy(iy)
          ix = ix + incx
          iy = iy + incy
   10   continue
        ddot = dtemp
        return
c
c       code for both increments equal to 1
c
   20   continue
        do 30 i = 1,n
          dtemp = dtemp + dx(i)*dy(i)
   30   continue
        ddot = dtemp
c
        return
        end function
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c Scales a vector by a constant.
c jack dongarra, linpack, 3/11/78.
c ----------------------------------------------------------------------
        subroutine dscal(n,da,dx,incx)
        double precision da,dx(1)
        integer i,incx,n,nincx
c
        if(n.le.0)return
        if(incx.eq.1)go to 20
c
c      code for increment not equal to 1
c
        nincx = n*incx
        do 10 i = 1,nincx,incx
         dx(i) = da*dx(i)
   10   continue
        return
c
c	code for increment equal to 1
c
   20   continue
        do 30 i = 1,n
         dx(i) = da*dx(i)
   30   continue
c
        return
        end subroutine
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c Finds the index of element having max. dabsolute value.
c jack dongarra, linpack, 3/11/78.
c ----------------------------------------------------------------------
        integer function idamax(n,dx,incx)
        double precision dx(1),dmax
        integer i,incx,ix,n
c
        idamax = 0
        if( n .lt. 1 ) return
        idamax = 1
        if(n.eq.1)return
        if(incx.eq.1)go to 20
c
c       code for increment not equal to 1
c
        ix = 1
        dmax = dabs(dx(1))
        ix = ix + incx
        do 10 i = 2,n
          if(dabs(dx(ix)).le.dmax) go to 5
          idamax = i
          dmax = dabs(dx(ix))
    5     ix = ix + incx
   10   continue
        return
c
c       code for increment equal to 1
c
   20   dmax = dabs(dx(1))
        do 30 i = 2,n
          if(dabs(dx(i)).le.dmax) go to 30
          idamax = i
          dmax = dabs(dx(i))
   30   continue
c
        return
        end function
c +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c
c This program should function properly on all systems
c satisfying the following two assumptions,
c  1.  the base used in representing dfloating point
c      numbers is not a power of three.
c  2.  the quantity  a  in statement 10 is represented to 
c      the accuracy used in dfloating point variables
c      that are stored in memory.
c      the statement number 10 and the go to 10 are intended to
c      force optimizing compilers to generate code satisfying 
c      assumption 2.
c      under these assumptions, it should be true that,
c      a is not exactly equal to four-thirds,
c      b  has a zero for its last bit or digit,
c      c  is not exactly equal to one,
c      eps  measures the separation of 1.0 from
c      the next larger dfloating point number.
c      the developers of eispack would appreciate being informed
c      about any systems where these assumptions do not hold.
c
c       *****************************************************************
c       this routine is one of the auxiliary routines used by eispack iii
c       to avoid machine dependencies.
c       *****************************************************************
c
c       this version dated 4/6/83.
c ----------------------------------------------------------------------
        double precision function epslon (x)
        double precision x
c       estimate unit roundoff in quantities of size x.
        double precision a,b,c,eps
c
        a = 4.0d0/3.0d0
   10   b = a - 1.0d0
        c = b + b + b
        eps = dabs(c-1.0d0)
        if (eps .eq. 0.0d0) go to 10
        epslon = eps*dabs(x)
c
        return
        end function
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c
c purpose:
c   multiply matrix m times vector x and add the result to vector y.
c
c parameters:
c
c n1  integer, number of elements in vector y, and number of rows 
c     in  matrix m
c
c y   double precision(n1), vector of length n1 to which is added 
c     the product m*x
c
c n2  integer, number of elements in vector x, and number of columns
c     in matrix m
c
c ldm integer, leading dimension of array m
c
c x   double precision(n2), vector of length n2
c
c m   double precision(ldm,n2), matrix of n1 rows and n2 columns
c
c ----------------------------------------------------------------------
        subroutine dmxpy (n1, y, n2, ldm, x, m)
        double precision y(*), x(*), m(ldm,*)
c
c       cleanup odd vector
c
        j = mod(n2,2)
        if (j .ge. 1) then
          do 10 i = 1, n1
            y(i) = (y(i)) + x(j)*m(i,j)
   10     continue
        endif
c
c       cleanup odd group of two vectors
c
          j = mod(n2,4)
          if (j .ge. 2) then
            do 20 i = 1, n1
              y(i) = ( (y(i))
     $      + x(j-1)*m(i,j-1)) + x(j)*m(i,j)
   20     continue
        endif
c
c       cleanup odd group of four vectors
c
        j = mod(n2,8)
        if (j .ge. 4) then
          do 30 i = 1, n1
            y(i) = ((( (y(i))
     $    + x(j-3)*m(i,j-3)) + x(j-2)*m(i,j-2))
     $    + x(j-1)*m(i,j-1)) + x(j)  *m(i,j)
   30     continue
        endif
c
c       cleanup odd group of eight vectors
c
        j = mod(n2,16)
        if (j .ge. 8) then
          do 40 i = 1, n1
            y(i) = ((((((( (y(i))
     $    + x(j-7)*m(i,j-7)) + x(j-6)*m(i,j-6))
     $    + x(j-5)*m(i,j-5)) + x(j-4)*m(i,j-4))
     $    + x(j-3)*m(i,j-3)) + x(j-2)*m(i,j-2))
     $    + x(j-1)*m(i,j-1)) + x(j)  *m(i,j)
   40     continue
        endif
c
c       main loop - groups of sixteen vectors
c
        jmin = j+16
        do 60 j = jmin, n2, 16
          do 50 i = 1, n1
            y(i) = ((((((((((((((( (y(i))
     $    + x(j-15)*m(i,j-15)) + x(j-14)*m(i,j-14))
     $    + x(j-13)*m(i,j-13)) + x(j-12)*m(i,j-12))
     $    + x(j-11)*m(i,j-11)) + x(j-10)*m(i,j-10))
     $    + x(j- 9)*m(i,j- 9)) + x(j- 8)*m(i,j- 8))
     $    + x(j- 7)*m(i,j- 7)) + x(j- 6)*m(i,j- 6))
     $    + x(j- 5)*m(i,j- 5)) + x(j- 4)*m(i,j- 4))
     $    + x(j- 3)*m(i,j- 3)) + x(j- 2)*m(i,j- 2))
     $    + x(j- 1)*m(i,j- 1)) + x(j)   *m(i,j)
   50   continue
   60   continue
c
        return
        end subroutine
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        double precision function ran( iseed )
c
c  modified from the lapack auxiliary routine 10/12/92 jdelia
c -- lapack auxiliary routine (version 1.0) --
c  univ. of tennessee, univ. of california berkeley, nag ltd.,
c  courant institute, argonne national lab, and rice university
c  february 29, 1992
c
c  .. array arguments ..
        integer iseed( 4 )
c  ..
c
c  purpose
c  =======
c
c  dlaran returns a random real number from a uniform (0,1)
c  distribution.
c
c  arguments
c  =========
c
c  iseed (input/output) integer array, dimension (4)
c  on entry, the seed of the random number generator; the array
c  elements must be between 0 and 4095, and iseed(4) must be
c  odd.
c  on exit, the seed is updated.
c
c  further details
c  ===============
c
c  this routine uses a multiplicative congruential method with modulus
c  2**48 and multiplier 33952834046453 (see g.s.fishman,
c  'multiplicative congruential random number generators with modulus
c  2**b: an exhaustive analysis for b = 32 and a partial analysis for
c  b = 48', math. comp. 189, pp 331-344, 1990).
c
c  48-bit integers are stored in 4 integer array elements with 12 bits
c  per element. hence the routine is portable across machines with
c  integers of 32 bits or more.
c
c  ..   parameters ..
        integer m1, m2, m3, m4
        parameter ( m1 = 494, m2 = 322, m3 = 2508, m4 = 2549 )
        double precision one
        parameter ( one = 1.0d+0 )
        integer ipw2
        double precision r
        parameter ( ipw2 = 4096, r = one / ipw2 )
c  ..
c  ..   local scalars ..
        integer it1, it2, it3, it4
c  ..
c  ..   intrinsic functions ..
        intrinsic dble, mod
c  ..
c  ..   executable statements ..
c
c       multiply the seed by the multiplier modulo 2**48
c
        it4 = iseed( 4 )*m4
        it3 = it4 / ipw2
        it4 = it4 - ipw2*it3
        it3 = it3 + iseed( 3 )*m4 + iseed( 4 )*m3
        it2 = it3 / ipw2
        it3 = it3 - ipw2*it2
        it2 = it2 + iseed( 2 )*m4 + iseed( 3 )*m3 + iseed( 4 )*m2
        it1 = it2 / ipw2
        it2 = it2 - ipw2*it1
        it1 = it1 + iseed( 1 )*m4 + iseed( 2 )*m3 + iseed( 3 )*m2 +
     $  iseed( 4 )*m1
        it1 = mod( it1, ipw2 )
c
c       return updated seed
c
        iseed (1) = it1
        iseed (2) = it2
        iseed (3) = it3
        iseed (4) = it4
c
c       convert 48-bit integer to a real number in the interval (0,1)
c
        ran = r*(dble(it1)+r*(dble(it2)+r*(dble(it3)+r*(dble(it4)))))
c
        return
        end function
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        subroutine aloca(ip,len,iend,nw,name,idebug)
        implicit real*8 (a-h,o-z) 
        character*(*) name
        if (len.eq.0)
     *    stop ' 1 aloca: longeur de memoire demande nulle'
        ip = iend
        iendd = ip + len
        if (iendd.gt.nw) then
          write (6,100) len, iendd, nw 
          stop
        endif
        iend=iendd
        if (idebug.eq.1)
     *    write (*,*) ' 1 aloca: variable: ''',name,''' length: ', len,
     *    ' start: ',ip,' end: ',iendd-1
  100     format (' 1 raloca: memoire excede'/
     *    ' longueur demande =',i10/
     *    ' taille totale ocuppe = ',i10/
     *    ' taille totale de memoire = ',i10)
c
        return
        end subroutine
c ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
